/*
 * bulk_extractor.cpp:
 * Feature Extraction Framework...
 *
 */

#include "bulk_extractor.h"
#include "aftimer.h"
#include "image_process.h"
#include "threadpool.h"
#include "xml.h"

#include <dirent.h>
#include <ctype.h>
#include <fcntl.h>
#include <set>
#include <setjmp.h>
#include <vector>
#include <queue>
#include <unistd.h>
#include <ctype.h>

#ifdef HAVE_MCHECK
#include <mcheck.h>
#endif

#ifdef HAVE_SYS_RESOURCE_H
#include <sys/resource.h>
#endif

#ifdef HAVE_SYS_SYSCTL_H
#include <sys/sysctl.h>
#endif

be_config_t be_config; // system configuration

using namespace std;

/****************************************************************
 *** COMMAND LINE OPTIONS
 ****************************************************************/

const char *progname=0;

size_t opt_pagesize=1024*1024*16;	// 
size_t opt_margin = 1024*1024*4;
size_t opt_max_feature_size = 1024*1024*1024;
size_t opt_max_context_size = 1024*1024*1024;
int opt_notify_rate = 4;
int word_min = 6;
int word_max = 14;
int num_threads = 1;
int64_t opt_offset_start = 0;
int64_t opt_offset_end   = 0;
int   min_uncompr_size = 6;	// don't bother with objects smaller than this
int   max_uncompr_size = 256*1024*1024; // don't decompress objects larger than this
int   debug=0;
int   opt_quiet = 0;			// print each time through
int   opt_silent= 0;
int   max_bad_alloc_errors = 60;
int64_t opt_page_start = 0;
uint32_t   opt_last_year = 2020;
const char *image_fname = 0;
const time_t max_wait_time=3600;

std::string HTTP_EOL = "\r\n";		// stdout is in binary form

/* Obtain subversion keywords
 * http://svnbook.red-bean.com/en/1.4/svn.advanced.props.special.keywords.html
 * http://stackoverflow.com/questions/1449935/getting-svn-revision-number-into-a-program-automatically
 * remember to set the property 'keywords=Revision' on bulk_extractor.cpp
 */
std::string svn_date("$Date: 2012-09-18 08:11:02 -0400 (Tue, 18 Sep 2012) $");
std::string svn_revision("$Rev: 10606 $");
std::string svn_author("$Author: simsong $");
std::string svn_headurl("$HeadURL: https://domex.nps.edu/domex/svn/src/bulk_extractor/trunk/src/bulk_extractor.cpp $");
std::string svn_id("$Id: bulk_extractor.cpp 10606 2012-09-18 12:11:02Z simsong $");

std::string svn_revision_clean()
{
    string svn_r;
    for(string::const_iterator it = svn_revision.begin();it!=svn_revision.end();it++){
	if(*it>='0' && *it<='9') svn_r.push_back(*it);
    }
    return svn_r;
}

/**
 * Singletons:
 * histograms_t histograms - the set of histograms to make
 * feature_recorder_set fs - the collection of feature recorders.
 * xml xreport             - the DFXML output.
 * image_process p         - the image being processed.
 * 
 * Note that all of the singletons are passed to the phase1() function.
 */


#ifdef WIN32 
// Allows us to open standard input in binary mode by default 
// See http://gnuwin32.sourceforge.net/compile.html for more 
int _CRT_fmode = _O_BINARY;
#endif


/****************************************************************
 *** SCANNER PLUG-IN SYSTEM
 ****************************************************************/

/* scanner_def is the class that is used internally to track all of the plug-in scanners.
 * Some scanners are compiled-in; others can be loaded at run-time.
 * Scanners can have multiple feature recorders. Multiple scanners can record to the
 * same feature recorder.
 */

/* An array of the built-in scanners */
scanner_t *scanners_builtin[] = {
    scan_accts,
    scan_base16,
    scan_base64,
    scan_kml,
    scan_email,
    //    scan_httpheader,
    scan_gps,
    scan_net,
    scan_find,
    scan_wordlist,
    scan_aes,
    scan_json,
    //scan_lift,  // not ready for prime time
    //scan_extx,  // not ready for prime time
#ifdef HAVE_EXIV2
    scan_exiv2,
#endif
    scan_elf,
    scan_exif,
    scan_zip,
    scan_gzip,
    scan_pdf,
    scan_winpe,
    scan_hiberfile,
    scan_winprefetch,
    scan_windirs,
    scan_vcard,
    scan_bulk,
    0};


/****************************************************************
 *** PATH PRINTER
 ****************************************************************/

/* Get the next token from the path. Tokens are separated by dashes.*/
static string get_and_remove_token(string &path)
{
    while(path[0]=='-'){
	path = path.substr(1); // remove any leading dashes.
    }
    size_t dash = path.find('-');	// find next dash
    if(dash==string::npos){		// no string; return the rest
	string prefix = path;
	path = "";
	return prefix;
    }
    string prefix = path.substr(0,dash);
    path = path.substr(dash+1);
    return prefix;
}

void process_path_printer(const scanner_params &sp)
{
    /* 1. Get next token 
     * 2. if prefix part is a number, skip forward that much in sbuf and repeat.
     *    if the prefix is PRINT, print the buffer
     *    if next part is a string, strip it and run that decoder.
     *    if next part is a |, print
     */
    if(debug & DEBUG_PRINT_STEPS) cerr << "process_path_printer " << sp.sbuf.pos0.path << "\n";
    string new_path = sp.sbuf.pos0.path;
    string prefix = get_and_remove_token(new_path);

    /* Time to print ?*/
    if(prefix.size()==0 || prefix=="PRINT"){

	uint64_t print_start = 0;
	uint64_t print_len = 4096;
    
	/* Check for options */
	scanner_params::PrintOptions::iterator it;

	it = sp.print_options.find("Content-Length");
	if(it!=sp.print_options.end()){
	    print_len = stoi64(it->second);
	}

	it = sp.print_options.find("Range");
	if(it!=sp.print_options.end()){
	    if(it->second[5]=='='){
		size_t dash = it->second.find('-');
		string v1 = it->second.substr(6,dash-6);
		string v2 = it->second.substr(dash+1);
		print_start = stoi64(v1);
		print_len = stoi64(v2)-print_start+1;
	    }
	}

	if(print_start>sp.sbuf.bufsize){
	    print_len = 0;			// can't print anything
	}

	if(print_len>0 && print_start+print_len>sp.sbuf.bufsize){
	    print_len = sp.sbuf.bufsize-print_start;
	}

	switch(scanner_params::getPrintMode(sp.print_options)){
	case scanner_params::MODE_HTTP:
	    std::cout << "Content-Length: "		<< print_len  << HTTP_EOL;
	    std::cout << "Content-Range: bytes "	<< print_start << "-" << print_start+print_len-1 << HTTP_EOL;
	    std::cout << "X-Range-Available: bytes " << 0 << "-" << sp.sbuf.bufsize-1 << HTTP_EOL;
	    std::cout << HTTP_EOL;
	    sp.sbuf.raw_dump(std::cout,print_start,print_len); // send to stdout as binary
	    break;
	case scanner_params::MODE_RAW:
	    std::cout << print_len << HTTP_EOL;
	    std::cout.flush();
	    sp.sbuf.raw_dump(std::cout,print_start,print_len); // send to stdout as binary
	    break;
	case scanner_params::MODE_HEX:
	    sp.sbuf.hex_dump(std::cout,print_start,print_len);
	    break;
	case scanner_params::MODE_NONE:
	    break;
	}
	return;			// our job is done
    }
    /* If we are in an offset block, process recursively with the offset */
    if(isdigit(prefix[0])){
	uint64_t offset = stoi64(prefix);
	if(offset>sp.sbuf.bufsize){
	    printf("Error: %s only has %u bytes; can't offset to %u\n",
		   new_path.c_str(),(unsigned int)sp.sbuf.bufsize,(unsigned int)offset);
	    return;
	}
	process_path_printer(scanner_params(scanner_params::scan,
					    sbuf_t(new_path,sp.sbuf+offset),
					    sp.fs,sp.print_options));
	return;
    }
    /* Find the scanner and use it */
    for(scanner_vector::const_iterator it = current_scanners.begin();it!=current_scanners.end();it++){
	const string &name = upperstr((*it)->info.name);
	if(name==prefix){
	    ((*it)->scanner)(scanner_params(scanner_params::scan,
					    sbuf_t(new_path,sp.sbuf),
					    sp.fs,sp.print_options),
			     recursion_control_block(process_path_printer,name,true));
	    return;
	}
    }
    cerr << "Unknown name in path: " << prefix << "\n";
}

/**
 * process_path uses the scanners to decode the path for the purpose of
 * decoding the image data and extracting the information.
 */


static void process_path(const image_process &p,string path,scanner_params::PrintOptions &po)
{
    /* Check for "/r" in path which means print raw */
    if(path.size()>2 && path.substr(path.size()-2,2)=="/r"){
	path = path.substr(0,path.size()-2);
    }

    string prefix = get_and_remove_token(path);
    int64_t offset = stoi64(prefix);

    /* Get the offset. process */
    u_char *buf = (u_char *)calloc(opt_pagesize,1);
    if(!buf) errx(1,"Cannot allocate buffer");
    int count = p.pread(buf,opt_pagesize,offset);
    if(count<0){
	cerr << p.image_fname << ": " << strerror(errno) << " (Read Error)\n";
	return;
    }

    /* make up a bogus feature recorder set and feature recorder */
    feature_recorder_set fs(feature_recorder_set::DISABLED);

    pos0_t pos0(path+"-PRINT"); // insert the PRINT token
    sbuf_t sbuf(pos0,buf,count,count,true); // sbuf system will free
    scanner_params sp(scanner_params::scan,sbuf,fs,po);
    process_path_printer(sp);
}

/**
 * process a path for a given filename.
 * Opens the image and calls the function above.
 * Also implements HTTP server with "-http" option.
 */
static void process_path(const char *fn,string path)
{
    image_process *pp = image_process_open(fn,0);
    if(pp==0){
	if(path=="-http"){
	    std::cout << "HTTP/1.1 502 Filename " << fn << " is invalid" << HTTP_EOL << HTTP_EOL;
	} else {
	    cerr << "Filename " << fn << " is invalid\n";
	}
	exit(1);
    }

    if(path=="-"){
	/* process path interactively */
	printf("Path Interactive Mode:\n");
	if(pp==0){
	    cerr << "Invalid file name: " << fn << "\n";
	    exit(1);
	}
	do {
	    getline(cin,path);
	    if(path==".") break;
	    scanner_params::PrintOptions po;
	    scanner_params::setPrintMode(po,scanner_params::MODE_HEX);
	    process_path(*pp,path,po);
	} while(true);
	return;
    }
    if(path=="-http"){
	do {
	    /* get the HTTP query */
	    string line;		// the specific query
	    scanner_params::PrintOptions po;
	    scanner_params::setPrintMode(po,scanner_params::MODE_HTTP);	// options for this query
	    
	    getline(cin,line);
	    truncate_at(line,'\r');
	    if(line.substr(0,4)!="GET "){
		std::cout << "HTTP/1.1 501 Method not implemented" << HTTP_EOL << HTTP_EOL;
		return;
	    }
	    size_t space = line.find(" HTTP/1.1");
	    if(space==string::npos){
		std::cout << "HTTP/1.1 501 Only HTTP/1.1 is implemented" << HTTP_EOL << HTTP_EOL;
		return;
	    }
	    string p2 = line.substr(4,space-4);

	    /* Get the additional header options */
	    do {
		getline(cin,line);
		truncate_at(line,'\r');
		if(line.size()==0) break; // double new-line
		size_t colon = line.find(":");
		if(colon==string::npos){
		    std::cout << "HTTP/1.1 502 Malformed HTTP request" << HTTP_EOL;
		    return;
		}
		string name = line.substr(0,colon);
		string val  = line.substr(colon+1);
		while(val.size()>0 && (val[0]==' '||val[0]=='\t')) val = val.substr(1);
		po[name]=val;
	    } while(true);
	    /* Process some specific URLs */
	    if(p2=="/info"){
		std::cout << "X-Image-Size: " << pp->image_size() << HTTP_EOL;
		std::cout << "X-Image-Filename: " << pp->image_fname << HTTP_EOL;
		std::cout << "Content-Length: 0" << HTTP_EOL;
		std::cout << HTTP_EOL;
		continue;
	    }

	    /* Ready to go with path and options */
	    process_path(*pp,p2,po);
	} while(true);
	return;
    }
    scanner_params::PrintOptions po;
    scanner_params::print_mode_t mode = scanner_params::MODE_HEX;
    if(path.size()>2 && path.substr(path.size()-2,2)=="/r"){
	path = path.substr(0,path.size()-2);
	mode = scanner_params::MODE_RAW;
    }
    if(path.size()>2 && path.substr(path.size()-2,2)=="/h"){
	path = path.substr(0,path.size()-2);
	mode = scanner_params::MODE_HEX;
    }
    scanner_params::setPrintMode(po,mode);
    process_path(*pp,path,po);
}

/****************************************************************
 *** Phase 1 BUFFER PROCESSING
 ****************************************************************/


/** process_extract is the main workhorse. It is calls each scanner on each page. 
 * @param sp    - the scanner params, including the sbuf to process 
 * It is also the recursive entry point for sub-analysis.
 */

void process_extract(const class scanner_params &sp)
{
    const sbuf_t &sbuf = sp.sbuf;
    const pos0_t &pos0 = sp.sbuf.pos0;
    class feature_recorder_set &fs = sp.fs;

    if(debug & DEBUG_NO_SCANNERS) {
	cerr << "DEBUG_NO_SCANNERS sbuf:" << sbuf << "\n";
	return;
    }
    if(debug & DEBUG_PRINT_STEPS) {
	cerr << "process_extract(" << pos0 << "," << sbuf << ")\n";
    }
    if(debug & DEBUG_DUMP_DATA) {
	sbuf.hex_dump(cerr);
    }

    if(sp.depth >= scanner_def::max_depth){
	feature_recorder_set::alert_recorder->write(pos0,"process_extract: MAX DEPTH REACHED","");
	return;
    }

    /****************************************************************
     *** CALL EACH OF THE SCANNERS ON THE SBUF
     ****************************************************************/

    for(scanner_vector::iterator it = current_scanners.begin();it!=current_scanners.end();it++){
	if((*it)->enabled==false) continue;
	string name = (*it)->info.name;
	
	try {
	    if(debug & DEBUG_PRINT_STEPS){
		cerr << "calling scanner " << name << "...\n";
	    }
		    
	    /* Compute the effective path for stats */
	    bool inname=false;
	    string epath;
	    for(string::const_iterator cc=sp.sbuf.pos0.path.begin();cc!=sp.sbuf.pos0.path.end();cc++){
		if(isupper(*cc)) inname=true;
		if(inname) epath.push_back(toupper(*cc));
		if(*cc=='-') inname=false;
	    }
	    if(epath.size()>0) epath.push_back('-');
	    for(string::const_iterator cc=name.begin();cc!=name.end();cc++){
		epath.push_back(toupper(*cc));
	    }

	    /* Call the scanner.*/
	    aftimer t;
	    t.start();
	    recursion_control_block rcb(process_extract,upperstr(name),false);
	    ((*it)->scanner)(sp,rcb);
	    t.stop();
	    sp.fs.add_stats(epath,t.elapsed_seconds());
		    
	    if(debug & DEBUG_PRINT_STEPS){
		cerr << "     ... scanner " << name << " returned\n";
	    }
	}
	catch ( exception &e ) {
	    cerr << "Scanner  " << name << " Exception: " << e.what() << " processing " << sp.sbuf.pos0 << "\n";
	}
    }
    fs.flush_all();
}


/**
 * Sleep for a minimum of msec
 */
static void msleep(int msec)
{
#if _WIN32
    Sleep(msec);
    return;
#else
#  ifdef HAVE_USLEEP
    usleep(msec*1000);
#  else
    int sec = msec/1000;
    if(sec<1) sec=1;
    sleep(sec);			// posix
#  endif
#endif
}

/**
 * phase1 - For every page of the iterator, schedule work.
 */

class MyException : public std::exception {
public:;
    std::string s;
    MyException(const std::string &ss) : s(ss) {}
    ~MyException() throw() {};
    virtual const char* what() const throw() {
	return s.c_str();
    }
};


/* Status for SIGINFO */
threadpool *phase1_tp = 0;
static void print_tp_status()
{
    std::stringstream ss;
    if(!phase1_tp) return;
    for(int i=0;i<num_threads;i++){
	std::string status = phase1_tp->get_thread_status(i);
	if(status.size() && status!="Free"){
	    ss << "Thread " << i << ": " << status << "\n";
	}
    }
    std::cout << ss.str() << "\n";
}


static std::string minsec(time_t tsec)
{
    time_t min = tsec / 60;
    time_t sec = tsec % 60;
    std::stringstream ss;
    if(min>0) ss << min << " min ";
    if(sec>0) ss << sec << " sec";
    return ss.str();
	
}

static void phase1(image_process &p,feature_recorder_set &fs,
		   xml &xreport,
		   int64_t &page_number,int64_t &total_bytes,
		   xml::tagid_set_t &seen_page_ids,
		   aftimer &timer)
{
    md5_generator *md5g = new md5_generator();		// keep track of MD5
    uint64_t md5_next = 0;		// next byte to hash
    threadpool tp(num_threads,fs,xreport);
    
    phase1_tp = &tp;

    xreport.push("runtime","xmlns:debug=\"http://www.afflib.org/bulk_extractor/debug\"");
    if(opt_quiet==0) std::cout << "Phase 1.\n";

    for(image_process::iterator it = p.begin(); it!= p.end() ; ++it){
	if(opt_offset_end!=0 && it.raw_offset > opt_offset_end) break; // passed the offset
	if(page_number>=opt_page_start && it.raw_offset>=opt_offset_start){
	    if(seen_page_ids.find(it.get_pos0().str()) != seen_page_ids.end()){
		// We've seen this page; skip it.
		continue;
	    }

	    // attempt to get an sbuf. If we can't get it, we may be in a low-memory situation.
	    // wait for 30 seconds.

	    try {
		sbuf_t *sbuf = 0;
		for(int retry_count=0;retry_count<max_bad_alloc_errors && sbuf==0;retry_count++){
		    try {
			sbuf = it.sbuf_alloc(); // may throw exception
			if(sbuf==0) break;
		    }
		    catch (const std::bad_alloc &e) {
			// Low memory could come from a bad sbuf alloc or another low memory condition.
			// wait for a while and then try again...
			cerr << "Low Memory (bad_alloc) exception: " << e.what()
			     << " reading " << it.get_pos0()
			     << " (retry_count=" << retry_count << ")\n";
			std::stringstream ss;
			ss << "name='bad_alloc' " << "pos0='" << it.get_pos0() << "' "
			   << "retry_count='"     << retry_count << "' ";
			xreport.xmlout("debug:exception", e.what(), ss.str(), true);
			if(retry_count+1>=max_bad_alloc_errors){
			    cerr << "Too many errors encountered in a row. Diagnose and restart.\n";
			    exit(1);
			}
			cerr << "will wait for 60 seconds and try again...\n";
			msleep(60000);
		    }
		}
		if(sbuf==0) break;	// eof?
		sbuf->page_number = page_number;
		if(md5g){
		    if(sbuf->pos0.offset==md5_next){ 
			// next byte follows logically
			md5g->update(sbuf->buf,sbuf->pagesize);
			md5_next += sbuf->pagesize;
		    } else {
			// we had a logical gap; stop hashing
			delete md5g;	
			md5g = 0;
		    }
		}
		total_bytes += sbuf->pagesize;
		tp.schedule_work(sbuf);	// schedule the work
		
		/* Notify the user */
		if((opt_quiet==0) && (opt_notify_rate>0) && ((sbuf->page_number % opt_notify_rate)==0)){
		    time_t t = time(0);
		    struct tm tm;
		    localtime_r(&t,&tm);
		    printf("%2d:%02d:%02d %s (%4.2f%%) Done in %s at %s\n",
			   tm.tm_hour,tm.tm_min,tm.tm_sec,
			   it.str().c_str(),
			   it.fraction_done()*100.0,
			   timer.eta_text(it.fraction_done()).c_str(),
			   timer.eta_time(it.fraction_done()).c_str());
		}
	    }
	    catch (const std::exception &e) {
		// 
		std::stringstream ss;
		ss << "name='" << e.what() << "' " << "pos0='" << it.get_pos0() << "' ";
		cerr << "Exception " << e.what() << " skipping " << it.get_pos0() << "\n";
		xreport.xmlout("debug:exception", e.what(), ss.str(), true);
	    }
	}
	page_number++;
    }
	    
    if(!opt_quiet){
	std::cout << "All Data is Read; waiting for threads to finish...\n";
    }

    /* Now wait for all of the threads to be free */
    tp.mode = 1;			// waiting for workers to finish
    time_t wait_start = time(0);
    for(int32_t counter = 0;;counter++){
	int num_remaining = num_threads - tp.get_free_count();
	if(num_remaining==0) break;

	msleep(100);
	time_t time_waiting   = time(0) - wait_start;
	time_t time_remaining = max_wait_time - time_waiting;

	if(counter%60==0){
	    std::stringstream ss;
	    ss << "Time elapsed waiting for " << num_remaining
	       << " thread" << (num_remaining>1 ? "s" : "") 
	       << " to finish:\n    " << minsec(time_waiting) 
	       << " (please wait for another " << minsec(time_remaining) << ".)\n";
	    if(opt_quiet==0){
		std::cout << ss.str();
		if(counter>0) print_tp_status();
	    }
	    xreport.comment(ss.str());
	}
	if(time_waiting>max_wait_time){
	    std::cout << "\n\n";
	    std::cout << " ... this shouldn't take more than an hour. Exiting ... \n";
	    std::cout << " ... Please report to the bulk_extractor maintainer ... \n";
	    break;
	}
    }
    if(opt_quiet==0) std::cout << "All Threads Finished!\n";

    xreport.pop();			// pop runtime
    /* We can write out the source info now, since we (might) know the hash */
    xreport.push("source");
    xreport.xmlout("image_filename",p.image_fname);
    xreport.xmlout("image_size",p.image_size());  
    if(md5g){
	md5_t md5 = md5g->final();
	xreport.xmlout("hashdigest",md5.hexdigest(),"type='MD5'",false);
	delete md5g;
    }
    xreport.pop();			// source

    /* Record the feature files and their counts in the output */
    xreport.push("feature_files");
    for(feature_recorder_map::const_iterator it = tp.fs.frm.begin();
	it != tp.fs.frm.end(); it++){
	xreport.set_oneline(true);
	xreport.push("feature_file");
	xreport.xmlout("name",it->second->name);
	xreport.xmlout("count",it->second->count);
	xreport.pop();
	xreport.set_oneline(false);
    }
    std::cout << "Producer time spent waiting: " << tp.waiting.elapsed_seconds() << " sec.\n";
    
    xreport.xmlout("thread_wait",dtos(tp.waiting.elapsed_seconds()),"thread='0'",false);
    double worker_wait_average = 0;
    for(threadpool::worker_vector::const_iterator it=tp.workers.begin();it!=tp.workers.end();it++){
	worker_wait_average += (*it)->waiting.elapsed_seconds() / num_threads;
	std::stringstream ss;
	ss << "thread='" << (*it)->id << "'";
	xreport.xmlout("thread_wait",dtos((*it)->waiting.elapsed_seconds()),ss.str(),false);
    }
    xreport.pop();
    xreport.flush();
    std::cout << "Average consumer time spent waiting: " << worker_wait_average << " sec.\n";
    if(worker_wait_average > tp.waiting.elapsed_seconds()*2 && worker_wait_average>10){
	std::cout << "*******************************************\n";
	std::cout << "** bulk_extractor is probably I/O bound. **\n";
	std::cout << "**        Run with a faster drive        **\n";
	std::cout << "**      to get better performance.       **\n";
	std::cout << "*******************************************\n";
    }
    if(tp.waiting.elapsed_seconds() > worker_wait_average * 2 && tp.waiting.elapsed_seconds()>10){
	std::cout << "*******************************************\n";
	std::cout << "** bulk_extractor is probably CPU bound. **\n";
	std::cout << "**    Run on a computer with more cores  **\n";
	std::cout << "**      to get better performance.       **\n";
	std::cout << "*******************************************\n";
    }
    /* end of phase 1 */
}

/****************************************************************
 *** Usage
 ****************************************************************/

static void info_scanners(bool detailed)
{
    /* Print a list of scanners */
    load_scanners(scanners_builtin,histograms);
    std::cout << "\n";
    std::vector<std::string> enabled_wordlist;
    std::vector<std::string> disabled_wordlist;
    for(scanner_vector::const_iterator it = current_scanners.begin();it!=current_scanners.end();it++){
	if(detailed){
	    std::cout << "Scanner Name: " << (*it)->info.name << "\n";
	    std::cout << "flags:  ";
	    if((*it)->info.flags & scanner_info::SCANNER_DISABLED) std::cout << "SCANNER_DISABLED ";
	    if((*it)->info.flags & scanner_info::SCANNER_NO_USAGE) std::cout << "SCANNER_NO_USAGE ";
	    if((*it)->info.flags==0) std::cout << "NONE ";

	    std::cout << "\n";
	    std::cout << "Scanner Interface version: " << (*it)->info.si_version << "\n";
	    std::cout << "Author: " << (*it)->info.author << "\n";
	    std::cout << "Description: " << (*it)->info.description << "\n";
	    std::cout << "URL: " << (*it)->info.url << "\n";
	    std::cout << "Scanner Version: " << (*it)->info.scanner_version << "\n";
	    std::cout << "Feature Names: ";
	    for(set<string>::const_iterator i2 = (*it)->info.feature_names.begin();
		i2 != (*it)->info.feature_names.end();
		i2++){
		std::cout << *i2 << " ";
	    }
	    std::cout << "\n\n";
	}
	if((*it)->info.flags & scanner_info::SCANNER_NO_USAGE) continue;
	if((*it)->info.flags & scanner_info::SCANNER_DISABLED){
	    disabled_wordlist.push_back((*it)->info.name);
	} else {
	    enabled_wordlist.push_back((*it)->info.name);
	}
    }
    if(detailed) return;
    sort(disabled_wordlist.begin(),disabled_wordlist.end());
    sort(enabled_wordlist.begin(),enabled_wordlist.end());
    for(std::vector<std::string>::const_iterator it = disabled_wordlist.begin();it!=disabled_wordlist.end();it++){
	std::cout << "   -e " <<  *it << " - enable scanner " << *it << "\n";
    }
    std::cout << "\n";
    for(std::vector<std::string>::const_iterator it = enabled_wordlist.begin();it!=enabled_wordlist.end();it++){
	std::cout << "   -x " <<  *it << " - disable scanner " << *it << "\n";
    }
}

static void usage() __attribute__ ((__noreturn__));
static void usage()
{
    std::cout << "bulk_extractor version " PACKAGE_VERSION " " << svn_revision << "\n";
    std::cout << "Usage: " << progname << " [options] imagefile\n";
    std::cout << "  runs bulk extractor and outputs to stdout a summary of what was found where\n";
    std::cout << "\n";
    std::cout << "Required parameters:\n";
    std::cout << "   imagefile     - the file to extract\n";
    std::cout << " or  -R filedir  - recurse through a directory of files\n";
#ifdef HAVE_LIBEWF
    std::cout << "                  SUPPORT FOR E01 FILES COMPILED IN\n";
#endif
#ifdef HAVE_LIBAFFLIB
    std::cout << "                  SUPPORT FOR AFF FILES COMPILED IN\n";
#endif    
#ifdef HAVE_EXIV2
    std::cout << "                  EXIV2 COMPILED IN\n";
#endif    
    std::cout << "   -o outdir    - specifies output directory. Must not exist.\n";
    std::cout << "                  bulk_extractor creates this directory.\n";
    std::cout << "Options:\n";
    std::cout << "   -b banner.txt- Add banner.txt contents to the top of every output file.\n";
    std::cout << "   -r alert_list.txt  - a file containing the alert list of features to alert\n";
    std::cout << "                       (can be a feature file or a list of globs)\n";
    std::cout << "                       (can be repeated.)\n";
    std::cout << "   -w stop_list.txt   - a file containing the stop list of features (white list\n";
    std::cout << "                       (can be a feature file or a list of globs)s\n";
    std::cout << "                       (can be repeated.)\n";
    std::cout << "   -F <rfile>   - Read a list of regular expressions from <rfile> to find\n";
    std::cout << "   -f <regex>   - find occurrences of <regex>; may be repeated.\n";
    std::cout << "                  results go into find.txt\n";
    std::cout << "   -q nn        - Quiet Rate; only print every nn status reports. Default 0; -1 for no status at all\n";
    std::cout << "\nTuning parameters:\n";
    std::cout << "   -C NN         - specifies the size of the context window (default " << feature_recorder::context_window << ")\n";
    std::cout << "   -G NN         - specify the page size (default " << opt_pagesize << ")\n";
    std::cout << "   -g NN         - specify margin (default " <<opt_margin << ")\n";
    std::cout << "   -W n1:n2      - Specifies minimum and maximum word size\n";
    std::cout << "                  (default is -w" << word_min << ":" << word_max << ")\n";
    std::cout << "   -B NN         - Specify the blocksize for bulk data analysis (default " <<opt_scan_bulk_block_size<< ")\n";
    std::cout << "   -j NN         - Number of analysis threads to run (default " <<threadpool::numCPU() << ")\n";
    std::cout << "   -M nn        - sets max recursion depth (default " << scanner_def::max_depth << ")\n";
    std::cout << "\nPath Processing Mode:\n";
    std::cout << "   -p <path>/f  - print the value of <path> with a given format.\n";
    std::cout << "                  formats: r = raw; h = hex.\n";
    std::cout << "                  Specify -p - for interactive mode.\n";
    std::cout << "                  Specify -p -http for HTTP mode.\n";
    std::cout << "\nParallelizing:\n";
    std::cout << "   -Y <o1>      - Start processing at o1 (o1 may be 1, 1K, 1M or 1G)\n";
    std::cout << "   -Y <o1>-<o2> - Process o1-o2\n";
    std::cout << "   -A <off>     - Add <off> to all reported feature offsets\n";
    std::cout << "\nDebugging:\n";
    std::cout << "   -h           - print this message\n";
    std::cout << "   -H           - print detailed info on the scanners\n";
    std::cout << "   -V           - print version number\n";
    std::cout << "   -z nn        - start on page nn\n";
    std::cout << "   -dN          - debug mode (see source code\n";
    std::cout << "   -Z           - zap (erase) output directory\n";
    std::cout << "\nControl of Scanners:\n";
    std::cout << "   -P <dir>     - Specifies a plugin directory\n";
    std::cout << "   -E scanner   - turn off all scanners except scanner\n";
    std::cout << "   -m <max>     - maximum number of minutes to wait for memory starvation\n";
    std::cout << "                  default is " << max_bad_alloc_errors << "\n";
    std::cout << "   -s name=value - sets a bulk extractor option name to be value\n";

    info_scanners(false);
    std::cout << "\n";
    exit(1);
}


/*
 * Make sure that the filename provided is sane.
 * That is, do not allow the analysis of a *.E02 file...
 */
void validate_fn(const string &fn)
{
    int r= access(fn.c_str(),R_OK);
    if(r!=0){
	cerr << "cannot open: " << fn << ": " << strerror(errno) << " (code " << r << ")\n";
	exit(1);
    }
    if(fn.size()>3){
	size_t e = fn.rfind('.');
	if(e!=string::npos){
	    string ext = fn.substr(e+1);
	    if(ext=="E02" || ext=="e02"){
		cerr << "Error: invalid file name\n";
		cerr << "Do not use bulk_extractor to process individual EnCase files.\n";
		cerr << "Instead, just run bulk_extractor with FILENAME.E01\n";
		cerr << "The other files in an EnCase multi-volume archive will be opened\n";
		cerr << "automatically.\n";
		exit(1);
	    }
	}
    }
}

/**
 * Create the dfxml output
 */

static void dfxml_create(xml &xreport,const string &command_line)
{
    xreport.push("dfxml","xmloutputversion='1.0'");
    xreport.push("metadata",
		 "\n  xmlns='http://IAMHEREafflib.org/bulk_extractor/' "
		 "\n  xmlns:xsi='http://www.w3.org/2001/XMLSchema-instance' "
		 "\n  xmlns:dc='http://purl.org/dc/elements/1.1/'" );
    xreport.xmlout("dc:type","Feature Extraction","",false);
    xreport.pop();
    xreport.add_DFXML_creator(PACKAGE_NAME,PACKAGE_VERSION,svn_revision_clean(),command_line);
    xreport.push("configuration");
    xreport.xmlout("threads",num_threads);
    xreport.push("scanners");
    /* Generate a list of the scanners in use */

    for(scanner_vector::const_iterator it=current_scanners.begin();it!=current_scanners.end();it++){
	if((*it)->enabled){
	    xreport.xmlout("scanner",(*it)->info.name);
	}
    }
    xreport.pop();			// scanners
    xreport.pop();			// configuration
}

class scanner_command {
public:
    enum command_t {DISABLE_ALL=0,ENABLE,DISABLE};
    scanner_command(const scanner_command &sc):command(sc.command),name(sc.name){};
    scanner_command(scanner_command::command_t c,const string &n):command(c),name(n){};
    command_t command;
    string name;
};

bool directory_missing(const std::string &d)
{
    return access(d.c_str(),F_OK)<0;
}

bool directory_empty(const std::string &d)
{
    if(directory_missing(d)==false){
	std::string reportfn = d + "/report.xml";
	if(access(reportfn.c_str(),R_OK)<0) return true;
    }
    return false;
}

int main(int argc,char **argv)
{
#ifdef HAVE_MCHECK
    mtrace();
#endif
    
    if(getenv("BULK_EXTRACTOR_DEBUG")){
	debug = atoi(getenv("BULK_EXTRACTOR_DEBUG"));
    }
    progname = argv[0];
    int64_t page_number = 0;
    const char *opt_path = 0;
    int opt_recurse = 0;
    int opt_zap = 0;
    string opt_outdir;
    char *cc;
    setvbuf(stdout,0,_IONBF,0);		// don't buffer stdout
    std::string command_line = xml::make_command_line(argc,argv);
    num_threads = threadpool::numCPU();

#ifdef WIN32
    setmode(1,O_BINARY);		// make stdout binary
    threadpool::win32_init();
#endif
    /* look for usage first */
    if(argc==1 || (strcmp(argv[1],"-h")==0)){
	usage();
	exit(1);
    }

    vector<scanner_command> scanner_commands;

    /* Process options */
    int ch;
    while ((ch = getopt(argc, argv, "A:B:b:C:d:E:e:F:f:G:g:Hhj:M:m:o:P:p:q:Rr:s:VW:w:x:Y:z:Z")) != -1) {
	switch (ch) {
	case 'A': feature_recorder::offset_add  = stoi64(optarg);break;
	case 'B': opt_scan_bulk_block_size = atoi(optarg);break;
	case 'b': feature_recorder::banner_file = optarg; break;
	case 'C': feature_recorder::context_window = atoi(optarg);break;
	case 'd':
	    {
		int d = atoi(optarg);
		switch(d){
		case DEBUG_ALLOCATE_512MiB: 
		    if(calloc(1024*1024*512,1)){
		      cerr << "-d1002 -- Allocating 512MB of RAM; may be repeated\n";
		    } else {
		      cerr << "-d1002 -- CANNOT ALLOCATE MORE RAM\n";
		    }
		    break;
		default:
		    debug  = d;
		    break;
		}
	    }
	    break;
	case 'E':
	    scanner_commands.push_back(scanner_command(scanner_command::DISABLE_ALL,string("")));
	    scanner_commands.push_back(scanner_command(scanner_command::ENABLE,optarg));
	    break;
	case 'e':
	    scanner_commands.push_back(scanner_command(scanner_command::ENABLE,optarg));
	    break;
	case 'F': process_find_file(optarg); break;
	case 'f': add_find_pattern(optarg); break;
	case 'G': opt_pagesize = atoi(optarg); break;
	case 'g': opt_margin = atoi(optarg); break;
	case 'j': num_threads = atoi(optarg); break;
	case 'M': scanner_def::max_depth = atoi(optarg); break;
	case 'm': max_bad_alloc_errors = atoi(optarg); break;
	case 'o': opt_outdir = optarg;break;
	case 'P': load_scanner_directory(optarg,histograms);break;
	case 'p': opt_path = optarg; break;
        case 'q':
	    opt_notify_rate = atoi(optarg);
	    if(opt_notify_rate<0) opt_quiet = 1; // -q -1 turns off notifications
	    break;
	case 'r': if(alert_list.readfile(optarg)){
		err(1,"Cannot read alert list %s",optarg);
	    }
	    break;
	case 'R': opt_recurse = 1; break;
	case 's':
	    {
		std::vector<std::string> params = split(optarg,'=');
		if(params.size()!=2){
		    std::cerr << "Invalid paramter: " << optarg << "\n";
		    exit(1);
		}
		be_config[params[0]] = params[1];
		continue;
	    }
	case 'V': std::cout << "bulk_extractor " << PACKAGE_VERSION << "\n"; exit (1);
	case 'W':
	    cc = strchr(optarg,':');
	    if(!cc) err(1,"-W requires n1:n2");
	    cc[0] = 0;			// null terminate
	    word_min = atoi(optarg);
	    word_max = atoi(cc+1);
	    if(word_min>word_max) err(1,"word_min=%d word_max=%d\n",word_min,word_max);
	    break;
	case 'w': if(stop_list.readfile(optarg)){
		err(1,"Cannot read stop list %s",optarg);
	    }
	    break;
	case 'x':
	    scanner_commands.push_back(scanner_command(scanner_command::DISABLE,optarg));
	    break;
	case 'Y': {
	    string optargs = optarg;
	    size_t dash = optargs.find('-');
	    if(dash==string::npos){
		opt_offset_start = stoi64(optargs);
		std::cerr << "Starting scanning at byte offset " << opt_offset_start << "\n";
	    } else {
		opt_offset_start = stoi64(optargs.substr(0,dash-1));
		opt_offset_end   = stoi64(optargs.substr(dash+1));
		std::cerr << "Scanning from by offset " << opt_offset_start << " to " << opt_offset_end << "\n";
	    }
	    break;
	}
	    
	case 'z': opt_page_start = stoi64(optarg);break;
	case 'Z': opt_zap=true;break;
	case 'H':
	    info_scanners(true);
	    exit(1);
	case 'h': case '?':default:
	    usage();
	    break;
	}
    }
    argc -= optind;
    argv += optind;


    if(be_config["work_start_work_end"]=="NO"){
	opt_work_start_work_end=false;
    }

    /* Load all the scanners and enable the ones we care about */
    load_scanners(scanners_builtin,histograms);		  // and load the scanners
    if(find_list.size()>0){
	set_scanner_enabled("find",1);
    }
    for(vector<scanner_command>::const_iterator it=scanner_commands.begin();it!=scanner_commands.end();it++){
	switch((*it).command){
	case scanner_command::DISABLE_ALL: disable_all_scanners(); break;
	case scanner_command::ENABLE:  set_scanner_enabled((*it).name,1);break;
	case scanner_command::DISABLE: set_scanner_enabled((*it).name,0);break;
	}
    }

    if(opt_path){
	if(argc!=1) errx(1,"-p requires a single argument.");
	process_path(argv[0],opt_path);
	exit(0);
    }
    if(opt_outdir.size()==0) errx(1,"error: -o outdir must be specified");

    /* The zap option wipes the contents of a directory, useful for debugging */
    if(opt_zap){
	DIR *dirp = opendir(opt_outdir.c_str());
	if(dirp){
	    struct dirent *dp;
	    while ((dp = readdir(dirp)) != NULL){
		string name = dp->d_name;
		if(name=="." || name=="..") continue;
		string fname = opt_outdir + string("/") + name;
		unlink(fname.c_str());
		std::cout << "erasing " << fname << "\n";
	    }
	}
	if(rmdir(opt_outdir.c_str())){
	    cerr << "cannot rmdir " << opt_outdir << "\n";
	    exit(1);
	}
	std::cout << "rmdir " << opt_outdir << "\n";
    }

    /* Start the clock */
    aftimer timer;
    timer.start();

    /* If output directory does not exist, we are not restarting! */
    xml  *xreport=0;
    string reportfilename = opt_outdir + "/report.xml";

    xml::tagid_set_t seen_page_ids; // pages that do not need re-processing
    image_process *p = 0;
    image_fname = *argv;

    if(directory_missing(opt_outdir) || directory_empty(opt_outdir)){
	/* Create the new directory */
	if ( argc !=1 ) errx(1,"Disk image option not provided. Run with -h for help.");

	/* If disk image does not exist, we are in restart mode */
	validate_fn(image_fname);
	if(directory_missing(opt_outdir)) be_mkdir(opt_outdir);
	p = image_process_open(image_fname,opt_recurse);
	if(!p){
	    if(errno) std::cerr << "Cannot open " << image_fname << ": " << strerror(errno) << "\n";
	    exit(1);
	}

	/* Store the configuration in the XML file */
	xreport = new xml(reportfilename,false);
	dfxml_create(*xreport,command_line);
	
	xreport->xmlout("provided_filename",image_fname); // save this information
    } else {
	/* Restarting */
	if(access(reportfilename.c_str(),R_OK)){
	    cerr << opt_outdir << ": error\n";
	    cerr << "report.xml file is missing or unreadable.\n";
	    cerr << "Directory may not have been created by bulk_extractor.\n";
	    cerr << "Cannot continue.\n";
	    exit(1);
	}
	
	/* Specify the tags for which we are searching for */
	xml::tagmap_t tagmap;
	tagmap["provided_filename"] = "";
	string dws("debug:work_start");
	string attrib = "pos0";

	class xml::existing e;
	e.tagmap = &tagmap;
	e.tagid = &dws;
	e.attrib = &attrib;
	e.tagid_set = &seen_page_ids;

	xreport = new xml(reportfilename,e);

	if(tagmap["provided_filename"].size()==0){
	    cerr << "report.xml in output directory is incomplete.\n";
	    cerr << "cannot restart.\n";
	    exit(1);
	}
	
	if(tagmap["provided_filename"] != image_fname){
	    cerr << "report.xml in output directory is for " << tagmap["provided_filename"] << "\n";
	    cerr << "cannot restart with " << image_fname << "\n";
	    exit(1);
	}

	p = image_process_open(image_fname,opt_recurse);
	std::cout << "Restarting from " << opt_outdir << "\n";
    }

    if(p==0) err(1,"Cannot image_process_open(%s)",image_fname);
	    
    /* Determine the feature files that will be used */
    feature_file_names_t feature_file_names;
    enable_feature_recorders(feature_file_names);

    feature_recorder_set fs(feature_file_names,opt_outdir);
    feature_recorder_set::alert_recorder = fs.get_name(feature_recorder_set::ALERT_RECORDER);
    int64_t total_bytes = 0;

    /* provide documentation to the user; the DFXML information comes from elsewhere */
    if(!opt_quiet){
	std::cout << "bulk_extractor version: " << PACKAGE_VERSION << "\n";
#ifdef HAVE_GETHOSTNAME
	char hostname[1024];
	gethostname(hostname,sizeof(hostname));
	std::cout << "Hostname: " << hostname << "\n";
#endif
	std::cout << "Input file: " << image_fname << "\n";
	std::cout << "Output directory: " << opt_outdir << "\n";
	std::cout << "Disk Size: " << p->image_size() << "\n";
	std::cout << "Threads: " << num_threads << "\n";
    }

    /****************************************************************
     *** THIS IS IT! 
     ****************************************************************/

    phase1(*p,fs,*xreport,page_number,total_bytes,seen_page_ids,timer);
    phase2(fs,*xreport);
    phase3(fs,*xreport);

    /* report and then print final usage information */
    xreport->push("report");
    xreport->xmlout("total_bytes",total_bytes);
    xreport->xmlout("elapsed_seconds",timer.elapsed_seconds());
    xreport->pop();			// report
    xreport->flush();
    fs.dump_stats(*xreport);
    xreport->add_rusage();
    xreport->pop();			// bulk_extractor
    xreport->close();
    delete p;				// not strictly needed, but why not?
    if(!opt_quiet){
	float mb_per_sec = (total_bytes / 1000000.0) / timer.elapsed_seconds();

	std::cout.precision(4);
	std::cout << "Elapsed time: " << timer.elapsed_seconds() << " sec.\n";
	std::cout << "Overall performance: " << mb_per_sec << " MBytes/sec.\n";
	feature_recorder *fr = fs.get_name("email");
	if(fr){
	    std::cout << "Total " << fr->name << " features found: " << fr->count << "\n";
	}
    }
#ifdef HAVE_MCHECK
    muntrace();
#endif
    exit(0);
}

